from __future__ import annotations

import logging
from datetime import datetime as dt
from datetime import timezone as tz

try:
    from typing import Self, cast
except ImportError:
    from typing_extensions import Self, cast

import pyotp
from sqlalchemy import Column, DateTime, ForeignKey, Text
from sqlalchemy.orm import Mapped

import ckan.plugins.toolkit as tk
from ckan import model
from ckan.model.types import make_uuid

import ckanext.auth.config as auth_config
from ckanext.auth.exceptions import ReplayAttackError

log = logging.getLogger(__name__)


class UserSecret(tk.BaseModel):
    __tablename__ = "2fa_user_secret"

    id: Mapped[str] = Column(Text, primary_key=True, default=make_uuid)  # type: ignore
    user_id: Mapped[str] = Column(ForeignKey(model.User.id, ondelete="CASCADE"), primary_key=True)  # type: ignore
    secret: Mapped[str] = Column(Text, nullable=False)  # type: ignore
    last_access: Mapped[dt | None] = Column(DateTime)  # type: ignore

    @classmethod
    def create_for_user(cls, user_name: str) -> Self:
        """Creates a new security challenge for the user.

        Args:
            user_name (str | None): user name

        Raises:
            ValueError: if the user_name is not provided

        Returns:
            Self: the new security challenge
        """
        secret_value = pyotp.random_base32()
        user_secret = cls.get_for_user(user_name)

        user = model.Session.query(model.User).filter(model.User.name == user_name).first()

        if not user:
            raise tk.ObjectNotFound("User not found")

        if not user_secret:
            user_secret = cls(user_id=user.id, secret=secret_value)
        else:
            user_secret.secret = secret_value

        model.Session.add(user_secret)
        model.Session.commit()

        return user_secret

    @classmethod
    def get_for_user(cls, user_name: str) -> Self | None:
        """Finds a secret object using the user name.

        :raises ValueError if the user_name is not provided
        """
        user = model.User.get(user_name)

        if not user:
            raise tk.ObjectNotFound("User not found")

        return (
            model.Session.query(cls)
            .join(model.User, model.User.id == cls.user_id)
            .filter(model.User.id == user.id)
            .first()
        )

    def get_code(self) -> str:
        """Get the current code for the secret."""
        return pyotp.TOTP(
            cast(str, self.secret),
            interval=auth_config.get_2fa_email_interval(),
        ).now()

    def check_code(self, code: str, verify_only: bool = False) -> bool:
        """Check the code against the secret.

        Args:
            code (str): the code to check
            verify_only (bool, optional):
                if True, the code is not saved as a successful challenge.
                Defaults to False.

        Raises:
            ReplayAttackError: if the code has already been used

        Returns:
            bool: True if the code is valid, False otherwise
        """
        code = code.strip()
        is_totp_enabled = auth_config.is_totp_2fa_enabled()

        if is_totp_enabled:
            totp = pyotp.TOTP(cast(str, self.secret))
            # valid_window – extends the validity to this many counter ticks
            # before and after the current one
            result = totp.verify(code, valid_window=1)
        else:
            totp = pyotp.TOTP(
                cast(str, self.secret),
                interval=auth_config.get_2fa_email_interval(),
            )
            result = totp.verify(code)

        if result and not verify_only:
            # check for replay attack...
            if is_totp_enabled and self.last_access and totp.at(cast(dt, self.last_access)) == code:
                raise ReplayAttackError("The code has already been used")

            self.last_access = dt.now(tz.utc)
            model.Session.commit()
        else:
            log.debug("2FA: Failed to verify the totp code for user %s", self.user_id)

        return result

    @property
    def provisioning_uri(self):
        """Returns the uri for setting up a QR code."""
        user = model.Session.query(model.User).filter(model.User.id == self.user_id).first()

        if user is None:
            raise ValueError(
                f"No user found for UserSecret instance with user_id {self.user_id}",
            )

        return pyotp.TOTP(cast(str, self.secret)).provisioning_uri(
            user.name,
            issuer_name=tk.config["ckan.site_url"],
        )



from __future__ import annotations

from typing import cast
from urllib import parse

import pytest

import ckan.plugins.toolkit as tk

from ckanext.auth.model import UserSecret

CODE_LENGTH = 6


@pytest.mark.usefixtures("with_plugins", "clean_db")
class TestUserSecretModel:
    def test_get_secret_for_user_missing(self, user):
        assert not UserSecret.get_for_user(user["name"])

    def test_create_for_user(self, user):
        secret = UserSecret.create_for_user(user["name"])

        assert secret.user_id == user["id"]
        assert secret.secret
        assert secret.last_access is None

    def test_get_secret_for_user(self, user):
        """The secret can be retrieved by user name or user ID."""
        secret = UserSecret.create_for_user(user["name"])

        assert UserSecret.get_for_user(user["name"]) == secret
        assert UserSecret.get_for_user(user["id"]) == secret

    def test_calling_create_again_rotates_the_secret(self, user):
        secret = UserSecret.create_for_user(user["name"])
        old_secret = secret.secret
        secret = UserSecret.create_for_user(user["name"])
        assert secret.secret != old_secret

    def test_create_for_missing_user(self):
        with pytest.raises(tk.ObjectNotFound):
            UserSecret.create_for_user("missing")

    def test_get_code(self, user):
        secret = UserSecret.create_for_user(user["name"])

        code = secret.get_code()

        assert code
        assert len(code) == CODE_LENGTH
        assert code.isdigit()

    def test_check_code(self, user):
        secret = UserSecret.create_for_user(user["name"])
        code = secret.get_code()

        assert secret.check_code(code)
        assert not secret.check_code("invalid")

    def test_check_code_updated_last_access(self, user):
        secret = UserSecret.create_for_user(user["name"])
        code = secret.get_code()

        assert not secret.last_access
        secret.check_code(code)
        assert secret.last_access

    def test_check_code_verify_only_once(self, user):
        """We use it for test verify on the user 2MA configure page."""
        secret = UserSecret.create_for_user(user["name"])
        code = secret.get_code()

        assert not secret.last_access
        assert secret.check_code(code, verify_only=True)
        assert not secret.last_access

    def test_provisioning_uri(self, user):
        secret = UserSecret.create_for_user(user["name"])

        assert secret.provisioning_uri
        assert "otpauth://totp" in secret.provisioning_uri
        assert user["name"] in secret.provisioning_uri
        assert cast(str, secret.secret) in secret.provisioning_uri
        assert parse.quote_plus(tk.config["ckan.site_url"]) in secret.provisioning_uri



from __future__ import annotations

import logging
from typing import TypedDict

import ckan.plugins.toolkit as tk
from ckan import model, types
from ckan.lib import captcha
from ckan.logic import validate

import ckanext.auth.config as auth_config
from ckanext.auth import exceptions, utils
from ckanext.auth.logic import schema
from ckanext.auth.model import UserSecret

log = logging.getLogger(__name__)


class LoginResponse(TypedDict, total=False):
    success: bool
    error: str | None
    valid: bool | None


@validate(schema.auth_2fa_user_login)
def auth_2fa_user_login(
    context: types.Context,
    data_dict: types.DataDict,
) -> LoginResponse:
    tk.check_access("auth_2fa_user_login", context, data_dict)

    user = model.User.by_name(data_dict["login"])

    if not user:
        raise tk.ObjectNotFound("User not found")

    user_secret = UserSecret.get_for_user(user.name)

    if not user_secret:
        user_secret = UserSecret.create_for_user(user.name)

    if data_dict["mfa_type"] not in auth_config.ALLOWED_METHODS:
        raise ValueError("Invalid MFA type")

    try:
        success = user_secret.check_code(data_dict["code"])
    except exceptions.ReplayAttackError:
        return LoginResponse(
            success=False,
            error="The verification code has expired",
            valid=False,
        )

    return LoginResponse(
        success=success,
        error=("The verification code is invalid" if not success else None),
        valid=success,
    )


@validate(schema.auth_2fa_user_login)
def auth_2fa_check_credentials(
    context: types.Context,
    data_dict: types.DataDict,
) -> LoginResponse:
    tk.check_access("auth_2fa_user_login", context, data_dict)

    user = model.User.by_name(data_dict["login"])

    try:
        captcha.check_recaptcha(tk.request)
    except captcha.CaptchaError:
        log.info("2FA: Login failed for %s", data_dict["login"])
        return LoginResponse(success=False, error=tk._("Invalid reCAPTCHA"))

    if not user or not user.is_active or not user.validate_password(data_dict["password"]):
        log.info("2FA: Login failed for %s", data_dict["login"])

        utils.LoginManager.log_user_login_attempt(data_dict["login"])

        if utils.LoginManager.get_user_login_attempts(data_dict["login"]) > auth_config.get_2fa_max_attempts():
            utils.LoginManager.block_user_login(data_dict["login"])

        return LoginResponse(success=False, error=tk._("Invalid login or password"))

    if utils.LoginManager.is_login_blocked(data_dict["login"]):
        log.info("2FA: User %s is blocked from logging in", data_dict["login"])
        return LoginResponse(success=False, error=tk._("Too many login attempts"))

    return LoginResponse(success=True, error=None)



from __future__ import annotations

from ckanext.auth import config as auth_config


def is_totp_2fa_enabled() -> bool:
    return auth_config.is_totp_2fa_enabled()


def is_2fa_enabled() -> bool:
    return auth_config.is_2fa_enabled()


def get_2fa_method() -> str:
    return auth_config.get_2fa_method()


def is_2fa_dev_mode_enabled() -> bool:
    return auth_config.is_2fa_dev_mode()



from __future__ import annotations

from typing import cast
from unittest import mock

import pytest

from ckan import model

from ckanext.auth import utils

CODE_LENGTH = 6
SECRET_LENGTH = 32


@pytest.mark.usefixtures("with_plugins", "clean_db", "with_request_context")
class TestSendVerificationCodeToUser:
    @mock.patch("ckan.lib.mailer.mail_user")
    def test_send_verification_code_to_user(self, mocker, app, user):
        mocker.return_value = True
        assert utils.send_verification_email_to_user(user["id"])

    def test_user_does_not_exist(self):
        assert not utils.send_verification_email_to_user("xxx")


@pytest.mark.usefixtures("with_plugins", "clean_db")
class TestGetEmailVerificationCode:
    def test_get_email_verification_code(self, user):
        user_obj = cast(model.User, model.User.get(user["id"]))
        code = utils.get_email_verification_code(user_obj)

        assert code
        assert isinstance(code, str)
        assert len(code) == CODE_LENGTH
        assert code.isdigit()


@pytest.mark.usefixtures("with_plugins", "clean_db")
class TestRegenerateUserSecret:
    def test_regenerate_user_secret(self, user):
        secret = utils.regenerate_user_secret(user["id"])

        assert secret
        assert isinstance(secret, str)
        assert len(secret) == SECRET_LENGTH
        assert secret.isalnum()


@pytest.mark.usefixtures("with_plugins", "clean_db", "clean_redis")
class TestLoginManager:
    def test_no_login_attemps(self, user):
        assert utils.LoginManager.get_user_login_attempts(user["id"]) == 0

    def test_one_login_attempt(self, user):
        utils.LoginManager.log_user_login_attempt(user["id"])
        assert utils.LoginManager.get_user_login_attempts(user["id"]) == 1

    def test_block_user_login(self, user):
        utils.LoginManager.block_user_login(user["id"])
        assert utils.LoginManager.is_login_blocked(user["id"])

    def test_reset_for_user(self, user):
        utils.LoginManager.block_user_login(user["id"])
        assert utils.LoginManager.is_login_blocked(user["id"])

        utils.LoginManager.reset_for_user(user["id"])
        assert not utils.LoginManager.is_login_blocked(user["id"])



from __future__ import annotations

import logging

import click

from ckanext.auth import model, utils

logger = logging.getLogger(__name__)

__all__ = [
    "auth",
]


@click.group()
def auth():
    pass


@auth.command()
@click.argument("username")
def reset_secret(username: str) -> None:
    utils.regenerate_user_secret(username)


@auth.command()
@click.argument("username")
def unblock_user(username: str) -> None:
    utils.LoginManager.reset_for_user(username)


@auth.command()
def unblock_all() -> None:
    utils.LoginManager.reset_all()


@auth.command()
@click.argument("username")
def get_code(username: str) -> None:
    secret = model.UserSecret.get_for_user(username)

    if not secret:
        secret = model.UserSecret.create_for_user(username)

    click.secho(f"Secret: {secret.get_code()}", fg="green")



from __future__ import annotations

import logging
from typing import Any

from flask import Response

import ckan.plugins.toolkit as tk
from ckan import plugins as p
from ckan import types
from ckan.common import CKANConfig

from ckanext.auth import utils

log = logging.getLogger(__name__)


@tk.blanket.actions
@tk.blanket.auth_functions
@tk.blanket.helpers
@tk.blanket.blueprints
@tk.blanket.config_declarations
@tk.blanket.cli
class AuthPlugin(p.SingletonPlugin):
    p.implements(p.IConfigurer)
    p.implements(p.ISignal)
    p.implements(p.IAuthenticator, inherit=True)

    # IConfigurer

    def update_config(self, config_: CKANConfig) -> None:
        tk.add_template_directory(config_, "templates")
        tk.add_resource("assets", "auth")

    # ISignal

    def get_signal_subscriptions(self) -> types.SignalMapping:
        return {
            tk.signals.ckanext.signal("ap_main:collect_config_sections"): [
                self.collect_config_sections_subs,
            ],
            tk.signals.ckanext.signal("ap_main:collect_config_schemas"): [
                self.collect_config_schemas_subs,
            ],
        }

    @staticmethod
    def collect_config_sections_subs(sender: None) -> dict[str, Any]:
        return {
            "name": "Auth",
            "configs": [
                {
                    "name": "Configuration",
                    "blueprint": "auth_admin.config",
                    "info": "Auth settings",
                },
            ],
        }

    @staticmethod
    def collect_config_schemas_subs(sender: None) -> list[str]:
        return ["ckanext.auth:config_schema.yaml"]

    # IAuthenticator

    def login(self) -> str | Response:
        return utils.login()



from __future__ import annotations

import ckan.plugins.toolkit as tk

CONF_2FA_ENABLED = "ckanext.auth.2fa_enabled"
CONF_2FA_METHOD = "ckanext.auth.2fa_method"
CONF_2FA_SUBJECT = "ckanext.auth.2fa_subject"
CONF_2FA_EMAIL_INTERVAL = "ckanext.auth.2fa_email_interval"
CONF_2FA_LOGIN_TIMEOUT = "ckanext.auth.2fa_login_timeout"
CONF_2FA_MAX_ATTEMPTS = "ckanext.auth.2fa_login_max_attempts"
CONF_2FA_DEV_MODE = "ckanext.auth.2fa_dev_mode"

METHOD_EMAIL = "email"
METHOD_AUTHENTICATOR = "authenticator"
ALLOWED_METHODS = [METHOD_EMAIL, METHOD_AUTHENTICATOR]


def is_2fa_enabled() -> bool:
    return tk.asbool(tk.config[CONF_2FA_ENABLED])


def get_2fa_method() -> str:
    return tk.config[CONF_2FA_METHOD]


def get_2fa_subject() -> str:
    return tk.config[CONF_2FA_SUBJECT]


def is_email_2fa_enabled() -> bool:
    return get_2fa_method() == METHOD_EMAIL


def is_totp_2fa_enabled() -> bool:
    return get_2fa_method() == METHOD_AUTHENTICATOR


def get_2fa_email_interval() -> int:
    return int(tk.config[CONF_2FA_EMAIL_INTERVAL])


def get_2fa_login_timeout() -> int:
    return int(tk.config[CONF_2FA_LOGIN_TIMEOUT])


def get_2fa_max_attempts() -> int:
    return int(tk.config[CONF_2FA_MAX_ATTEMPTS])


def is_2fa_dev_mode() -> bool:
    return tk.asbool(tk.config[CONF_2FA_DEV_MODE])



from __future__ import annotations

import logging
from collections.abc import Callable
from functools import wraps
from typing import Any, cast

from flask import Blueprint, Response, jsonify, request
from flask.views import MethodView

from ckan import model, types
from ckan.lib import helpers
from ckan.logic import parse_params
from ckan.plugins import plugin_loaded
from ckan.plugins import toolkit as tk

import ckanext.auth.helpers as ah
from ckanext.auth import utils
from ckanext.auth.model import UserSecret

log = logging.getLogger(__name__)
auth = Blueprint("auth", __name__, url_prefix="/mfa")


def require_login(func: Callable[..., Response | str]) -> Callable[..., Response | str]:
    @wraps(func)
    def decorated_view(*args: tuple, **kwargs: dict[str, str]) -> Response | str:
        if tk.current_user.is_anonymous:
            return tk.abort(401, tk._("You have to be logged in to access this page."))
        return func(*args, **kwargs)

    return decorated_view


class Configure2FA(MethodView):
    @require_login
    def get(self, user_id: str) -> str:
        try:
            user_dict = tk.get_action("user_show")(self.get_context(), {"id:": user_id})
        except tk.ObjectNotFound:
            tk.abort(404, tk._("User not found"))
        except tk.NotAuthorized:
            tk.abort(403, tk._("Not authorized to see this page"))

        extra_vars = {
            "user_id": user_id,
            "is_sysadmin": user_dict["sysadmin"],
            "is_myself": user_dict["name"] == tk.current_user.name,
            "user_dict": user_dict,
        }

        extra_vars.update(self._setup_totp_extra_vars(user_id))

        return tk.render(
            "auth/configure_2fa.html",
            extra_vars,
        )

    def get_context(self) -> types.Context:
        return {
            "user": tk.current_user.name,
            "auth_user_obj": tk.current_user,
        }

    @require_login
    def post(self, user_id: str) -> Response:
        extra_vars = self._setup_totp_extra_vars(user_id)

        if extra_vars.get("code_valid"):
            tk.h.flash_success(
                tk._(
                    """
                    The code is valid. Your authenticator app is
                    now properly set up for future use.""",
                ),
            )
        else:
            tk.h.flash_error(
                tk._(
                    """The code is incorrect. Please try scanning
                the QR code with your authenticator app again.""",
                ),
            )

        return tk.redirect_to("auth.configure_2fa", user_id=user_id)

    def _setup_totp_extra_vars(self, user_id: str) -> dict[str, Any]:
        data_dict = parse_params(tk.request.form)

        user_secret = UserSecret.get_for_user(user_id)

        if not user_secret:
            return {}

        test_code = cast(str, data_dict.get("code"))

        extra_vars: dict[str, Any] = {
            "totp_secret": user_secret.secret,
            "provisioning_uri": user_secret.provisioning_uri,
        }

        if request.method == "POST" and test_code:
            extra_vars["code_valid"] = user_secret.check_code(
                test_code,
                verify_only=True,
            )

        return extra_vars


@auth.route("/configure_mfa/<user_id>/new", methods=["GET", "POST"])
@require_login
def regenerate_secret(user_id: str):
    utils.regenerate_user_secret(user_id)
    tk.h.flash_success(tk._("Your 2FA secret has been regenerated."))
    return helpers.redirect_to("auth.configure_2fa", user_id=user_id)


@auth.route("/send_verification_code", methods=["POST"])
def send_verification_code() -> Response:
    user_name: str = tk.get_or_bust(dict(tk.request.form), "login")

    utils.regenerate_user_secret(user_name)
    success = utils.send_verification_email_to_user(user_name)

    return jsonify(
        {
            "success": success,
            "error": ("Failed to send verification code" if not success else None),
            "result": None,
        },
    )


@auth.route("/init_qr_code", methods=["POST"])
def init_qr_code() -> Response:
    user_name: str = tk.get_or_bust(dict(tk.request.form), "login")

    secret = UserSecret.get_for_user(user_name)

    if not secret:
        secret = UserSecret.create_for_user(user_name)

    return jsonify(
        {
            "success": True,
            "error": None,
            "result": {
                "accessed": bool(secret.last_access),
                "provisioning_uri": secret.provisioning_uri,
                "secret": secret.secret,
            },
        },
    )


@auth.route("/get-user-code", methods=["POST"])
def get_2fa_user_code() -> Response:
    if not ah.is_2fa_dev_mode_enabled():
        tk.abort(404, tk._("Not found"))

    user_name = tk.request.form.get("login")

    if not user_name:
        return tk.abort(400, tk._("Missing 'login' parameter"))

    user = model.User.get(user_name)

    if not user or not user.email:
        return tk.abort(404, tk._("User not found or has no email set"))

    code = utils.get_email_verification_code(user)

    log.info("Providing 2FA code %s for user %s in dev mode", code, user.name)

    return jsonify({"success": True, "result": {"code": code}})


if plugin_loaded("admin_panel"):
    from ckanext.ap_main.utils import ap_before_request
    from ckanext.ap_main.views.generics import ApConfigurationPageView

    auth_admin = Blueprint("auth_admin", __name__)
    auth_admin.before_request(ap_before_request)

    auth_admin.add_url_rule(
        "/admin-panel/auth/config",
        view_func=ApConfigurationPageView.as_view(
            "config",
            "ckanext_auth_config",
            page_title=tk._("Auth config"),
        ),
    )


auth.add_url_rule(
    "/configure_2fa/<user_id>",
    view_func=Configure2FA.as_view("configure_2fa"),
)


def get_blueprints():
    return [auth]



from __future__ import annotations

from ckan import types
from ckan.logic.schema import validator_args

from ckanext.auth import config as auth_config


@validator_args
def auth_2fa_user_login(
    not_missing: types.Validator,
    unicode_safe: types.Validator,
    one_of: types.ValidatorFactory,
) -> types.Schema:
    return {
        "login": [not_missing, unicode_safe],
        "password": [not_missing, unicode_safe],
        "code": [not_missing, unicode_safe],
        "mfa_type": [
            not_missing,
            unicode_safe,
            one_of(auth_config.ALLOWED_METHODS),
        ],
    }



from __future__ import annotations

import pytest

from ckanext.auth import config, helpers


class TestIs2FAEnabled:
    def test_enabled_by_default(self):
        assert helpers.is_2fa_enabled()

    @pytest.mark.ckan_config(config.CONF_2FA_ENABLED, False)
    def test_disabled(self):
        assert not helpers.is_2fa_enabled()


class TestGet2FAMethod:
    @pytest.mark.ckan_config(config.CONF_2FA_METHOD, config.METHOD_AUTHENTICATOR)
    def test_set_authenticator_method(self):
        assert helpers.get_2fa_method() == config.METHOD_AUTHENTICATOR

    @pytest.mark.ckan_config(config.CONF_2FA_METHOD, config.METHOD_EMAIL)
    def test_set_email_method(self):
        assert helpers.get_2fa_method() == config.METHOD_EMAIL

    def test_default_is_email(self):
        assert helpers.get_2fa_method() == config.METHOD_EMAIL


class TestIsTOTP2FAEnabled:
    def test_not_enabled_by_default(self):
        assert not helpers.is_totp_2fa_enabled()

    @pytest.mark.ckan_config(config.CONF_2FA_METHOD, config.METHOD_AUTHENTICATOR)
    def test_enabled(self):
        assert helpers.is_totp_2fa_enabled()


class TestIsDevModeEnabled:
    def test_not_enabled_by_default(self):
        assert not helpers.is_2fa_dev_mode_enabled()

    @pytest.mark.ckan_config(config.CONF_2FA_DEV_MODE, True)
    def test_enabled(self):
        assert helpers.is_2fa_dev_mode_enabled()



from __future__ import annotations

import logging
from datetime import timedelta

try:
    from typing import NotRequired, TypedDict, cast
except ImportError:
    from typing_extensions import NotRequired, TypedDict, cast

import ckan.lib.mailer as ckan_mailer
import ckan.plugins as p
import ckan.plugins.toolkit as tk
from ckan import model
from ckan.lib.authenticator import default_authenticate
from ckan.lib.redis import connect_to_redis
from ckan.views.user import next_page_or_default, rotate_token

from ckanext.auth import config
from ckanext.auth import config as auth_config
from ckanext.auth.exceptions import ReplayAttackError
from ckanext.auth.model import UserSecret

log = logging.getLogger(__name__)


class IdentityDict(TypedDict):
    login: NotRequired[str]
    password: str
    check_captcha: bool


class LoginManager:
    login_attempts_key = "ckanext-auth:login_attempts:{}"
    blocked_key = "ckanext-auth:blocked:{}"

    @classmethod
    def is_login_blocked(cls, user_id: str) -> bool:
        """Check if a user is blocked from logging in."""
        return bool(connect_to_redis().get(cls.blocked_key.format(user_id)))

    @classmethod
    def block_user_login(cls, user_id: str) -> None:
        """Block a user from logging in for a certain amount of time."""
        connect_to_redis().setex(
            cls.blocked_key.format(user_id),
            auth_config.get_2fa_login_timeout(),
            1,
        )

    @classmethod
    def log_user_login_attempt(cls, user_id: str) -> None:
        """Log a login attempt for a user."""
        redis = connect_to_redis()

        redis.incr(cls.login_attempts_key.format(user_id))

    @classmethod
    def get_user_login_attempts(cls, user_id: str) -> int:
        """Get the number of login attempts for a user."""
        return int(connect_to_redis().get(cls.login_attempts_key.format(user_id)) or 0)  # type: ignore

    @classmethod
    def reset_for_user(cls, user_id: str) -> None:
        """Reset the login attempts for a user."""
        log.debug("2FA: Resetting login attempts for user %s", user_id)

        redis = connect_to_redis()

        redis.delete(cls.login_attempts_key.format(user_id))
        redis.delete(cls.blocked_key.format(user_id))

    @classmethod
    def reset_all(cls) -> None:
        """Reset the login attempts for all users."""
        log.debug("2FA: Resetting login attempts for all users")

        redis = connect_to_redis()

        for key in redis.keys(cls.login_attempts_key.format("*")):  # type: ignore
            redis.delete(key)

        for key in redis.keys(cls.blocked_key.format("*")):  # type: ignore
            redis.delete(key)


def send_verification_email_to_user(user_id: str) -> bool:
    user = model.User.get(user_id)

    if not user or not user.email:
        return False

    code = get_email_verification_code(user)
    data = {
        "verification_code": code,
        "site_url": tk.config["ckan.site_url"],
        "site_title": tk.config["ckan.site_title"],
        "user_name": user.display_name,
        "subject": tk._(config.get_2fa_subject()),
        "body": f"Your verification code is: {code}",
    }

    if p.plugin_loaded("mailcraft"):
        from ckanext.mailcraft.utils import get_mailer  # noqa PLC0415

        get_mailer().mail_recipients(
            subject=data["subject"],
            recipients=[user.email],
            body=data["body"],
            body_html=tk.render(
                "auth/emails/verification_code.html",
                extra_vars=data,
            ),
        )
    else:
        try:
            ckan_mailer.mail_user(
                recipient=user,
                subject=data["subject"],
                body=data["body"],
                body_html=tk.render(
                    "auth/emails/verification_code.html",
                    extra_vars=data,
                ),
            )
        except ckan_mailer.MailerException:
            return False

    return True


def get_email_verification_code(user: model.User) -> str:
    user_secret = UserSecret.get_for_user(user.name)

    if not user_secret:
        user_secret = UserSecret.create_for_user(user.name)

    return user_secret.get_code()


def regenerate_user_secret(user_id: str) -> str:
    """Regenerate the secret for a user.

    Args:
        user_id (str): The id of the user

    Returns:
        str: The new secret
    """
    user = model.User.get(user_id)

    if not user:
        raise tk.ObjectNotFound("User not found")

    user_secret = UserSecret.create_for_user(user.name)

    log.debug("2FA: Rotated the 2fa secret for user %s", user_id)

    return cast(str, user_secret.secret)


def login():
    if tk.current_user.is_authenticated:
        return tk.render("user/logout_first.html", {})

    if tk.request.method != "POST":
        return tk.render("user/login.html", {})

    user_obj = authenticate(
        IdentityDict(
            login=tk.get_or_bust(tk.request.form, "login"),
            password=tk.get_or_bust(tk.request.form, "password"),
            check_captcha=False,
        )
    )

    if not user_obj:
        tk.h.flash_error(tk._("Login failed. Bad username or password."))
        return tk.render("user/login.html", {})

    if remember := tk.request.form.get("remember"):
        tk.login_user(
            user_obj,
            remember=True,
            duration=timedelta(milliseconds=int(remember)),
        )
    else:
        tk.login_user(user_obj)

    rotate_token()

    return next_page_or_default(
        tk.request.args.get("next", tk.request.args.get("came_from")),
    )


def authenticate(identity: IdentityDict) -> model.User | model.AnonymousUser | None:
    # Run through the CKAN auth sequence first, so we can hit the DB
    # in every case and make timing attacks a little more difficult.
    ckan_auth_result = default_authenticate(dict(identity))

    if "login" not in identity:
        return None

    if LoginManager.is_login_blocked(identity["login"]):
        return None

    if LoginManager.get_user_login_attempts(identity["login"]) > config.get_2fa_max_attempts():
        LoginManager.block_user_login(identity["login"])

    if not ckan_auth_result:
        return LoginManager.log_user_login_attempt(identity["login"])

    if not config.is_2fa_enabled():
        LoginManager.reset_for_user(identity["login"])
        return ckan_auth_result

    # if the CKAN authenticator has successfully authenticated
    # then check the TOTP parameter to see if it is valid
    if authenticate_totp(identity["login"]):
        LoginManager.reset_for_user(identity["login"])
        return ckan_auth_result

    # This means that the login form has been submitted
    # with an invalid TOTP code, bypassing the ajax
    # login workflow.

    # The username and password were fine, but the 2fa
    # code was missing or invalid
    return None


def authenticate_totp(user_name: str) -> str | None:
    user_secret = UserSecret.get_for_user(user_name)

    # if there is no totp configured, don't allow auth
    # shouldn't happen, login flow should create a user secret
    if not user_secret:
        return log.debug(
            "2FA: Login attempted without MFA configured for: %s",
            user_name,
        )

    if "code" not in tk.request.form:
        return log.debug("2FA: Could not get MFA credentials from a request")

    try:
        result = user_secret.check_code(tk.request.form["code"])
    except ReplayAttackError as e:
        return log.warning(
            "2FA: Detected a possible replay attack for user: %s, context: %s",
            user_name,
            e,
        )
    else:
        return user_name if result else None



"""Init 2fa_user_totp table.

Revision ID: 7917e1c52a37
Revises:
Create Date: 2024-08-01 15:38:57.177385

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "7917e1c52a37"
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "2fa_user_secret",
        sa.Column("id", sa.Text(), nullable=False),
        sa.Column("user_id", sa.Text(), nullable=False),
        sa.Column("secret", sa.Text(), nullable=False),
        sa.Column("last_access", sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint("id", "user_id"),
    )


def downgrade():
    op.drop_table("2fa_user_secret")
